"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const src_1 = require("../src");
const util_1 = require("../src/util");
const art_1 = require("../src/art");
const test_utils_1 = require("./test.utils");
function serDeser(node) {
    const data = new Uint8Array(node.length * (1 + src_1.crypt.PK_LENGTH) + Array.from(node.leafs()).length * (2 * src_1.crypt.PK_LENGTH + src_1.crypt.SI_LENGTH));
    const offset1 = new util_1.Offset();
    (0, art_1.serializeTree)(data, new DataView(data.buffer), offset1, node);
    offset1.assert(data.length);
    const offset2 = new util_1.Offset();
    const de = (0, art_1.deserializeTree)(data, new DataView(data.buffer), offset2);
    offset2.assert(data.length);
    return de;
}
describe('ART Tree', () => {
    it('should throw with invalid data', () => {
        expect(() => (0, art_1.createTree)([], { identity: (0, util_1.keypair)(), exchange: (0, util_1.keypair)() })).toThrow();
        expect(() => (0, art_1.createTree)([null], { identity: (0, util_1.keypair)(), exchange: (0, util_1.keypair)() })).toThrow();
    });
    it('should construct 2-tree', () => {
        const [peer0, peer1] = (0, test_utils_1.generate)(2);
        if (!peer0 || !peer1)
            throw new Error();
        const exchange = (0, util_1.keypair)();
        const peers = [new art_1.Me(peer0.identity, peer0.ephemeral), new art_1.Peer(peer1.identity.pk, peer1.ephemeral.pk, peer1.ephemeral_signature)];
        const tree = (0, art_1.createTree)(peers, {
            identity: peer0.identity,
            exchange,
        });
        expect(tree.length).toEqual(3);
        expect(tree).toBeInstanceOf(art_1.Node);
        expect(tree.left).toBeInstanceOf(art_1.Leaf);
        expect(tree.right).toBeInstanceOf(art_1.Leaf);
        expect(tree.left.pk).toEqual(peer0.ephemeral.pk);
        expect(tree.right.pk).toEqual((0, util_1.keypair)((0, util_1.keyExchangeSend)(peer0.identity, peer1.identity.pk, exchange, peer1.ephemeral.pk)).pk);
        const reverse = serDeser(tree);
        expect(reverse.length).toEqual(3);
        expect(reverse).toBeInstanceOf(art_1.Node);
        expect(reverse.left).toBeInstanceOf(art_1.Leaf);
        expect(reverse.right).toBeInstanceOf(art_1.Leaf);
        expect(reverse.left.pk).toEqual(peer0.ephemeral.pk);
        expect(reverse.right.pk).toEqual((0, util_1.keypair)((0, util_1.keyExchangeReceive)(peer1.identity, peer0.identity.pk, peer1.ephemeral, exchange.pk)).pk);
        expect(reverse.pk).toEqual(tree.pk);
    });
    it('should construct 3-tree', () => {
        const [peer0, peer1, peer2] = (0, test_utils_1.generate)(3);
        if (!peer0 || !peer1 || !peer2)
            throw new Error();
        const exchange = (0, util_1.keypair)();
        const peers = [
            new art_1.Me(peer0.identity, peer0.ephemeral),
            new art_1.Peer(peer1.identity.pk, peer1.ephemeral.pk, peer1.ephemeral_signature),
            new art_1.Peer(peer2.identity.pk, peer2.ephemeral.pk, peer2.ephemeral_signature),
        ];
        const tree = (0, art_1.createTree)(peers, {
            identity: peer0.identity,
            exchange,
        });
        expect(tree.length).toEqual(5);
        expect(tree).toBeInstanceOf(art_1.Node);
        expect(tree.left).toBeInstanceOf(art_1.Node);
        expect(tree.right).toBeInstanceOf(art_1.Leaf);
        const n0pk = peer0.ephemeral.pk;
        expect(tree.left.left.pk).toEqual(n0pk);
        const n1 = (0, util_1.keypair)((0, util_1.keyExchangeSend)(peer0.identity, peer1.identity.pk, exchange, peer1.ephemeral.pk));
        expect(tree.left.right.pk).toEqual(n1.pk);
        const n2 = (0, util_1.keypair)((0, util_1.keyExchangeSend)(peer0.identity, peer2.identity.pk, exchange, peer2.ephemeral.pk));
        expect(tree.right.pk).toEqual(n2.pk);
        const n3 = (0, util_1.keypair)(src_1.crypt.hkdf(src_1.crypt.getSharedSecret(peer0.ephemeral.sk, n1.pk), undefined, undefined, src_1.crypt.SK_LENGTH));
        const n4 = (0, util_1.keypair)(src_1.crypt.hkdf(src_1.crypt.getSharedSecret(n3.sk, n2.pk), undefined, undefined, src_1.crypt.SK_LENGTH));
        expect(tree.pk).toEqual(n4.pk);
        const reverse = serDeser(tree);
        expect(reverse.length).toEqual(5);
        expect(reverse).toBeInstanceOf(art_1.Node);
        expect(reverse.left).toBeInstanceOf(art_1.Node);
        expect(reverse.right).toBeInstanceOf(art_1.Leaf);
        expect(reverse.left.left.pk).toEqual(tree.left.left.pk);
        expect(reverse.left.right.pk).toEqual(tree.left.right.pk);
        expect(reverse.right.pk).toEqual((0, util_1.keypair)((0, util_1.keyExchangeReceive)(peer2.identity, peer0.identity.pk, peer2.ephemeral, exchange.pk)).pk);
        expect(reverse.pk).toEqual(tree.pk);
    });
    it('should construct 4-tree', () => {
        const [peer0, peer1, peer2, peer3] = (0, test_utils_1.generate)(4);
        if (!peer0 || !peer1 || !peer2 || !peer3)
            throw new Error();
        const exchange = (0, util_1.keypair)();
        const peers = [
            new art_1.Me(peer0.identity, peer0.ephemeral),
            new art_1.Peer(peer1.identity.pk, peer1.ephemeral.pk, peer1.ephemeral_signature),
            new art_1.Peer(peer2.identity.pk, peer2.ephemeral.pk, peer2.ephemeral_signature),
            new art_1.Peer(peer3.identity.pk, peer3.ephemeral.pk, peer3.ephemeral_signature),
        ];
        const tree = (0, art_1.createTree)(peers, {
            identity: peer0.identity,
            exchange,
        });
        const dh1 = (0, util_1.keypair)((0, util_1.keyExchangeSend)(peer0.identity, peer1.identity.pk, exchange, peer1.ephemeral.pk));
        const dh2 = (0, util_1.keypair)((0, util_1.keyExchangeSend)(peer0.identity, peer2.identity.pk, exchange, peer2.ephemeral.pk));
        const dh3 = (0, util_1.keypair)((0, util_1.keyExchangeSend)(peer0.identity, peer3.identity.pk, exchange, peer3.ephemeral.pk));
        expect(tree.length).toEqual(7);
        expect(tree).toBeInstanceOf(art_1.Node);
        expect(tree.left).toBeInstanceOf(art_1.Node);
        expect(tree.right).toBeInstanceOf(art_1.Node);
        expect(tree.left.left.pk).toEqual(peer0.ephemeral.pk);
        expect(tree.left.right.pk).toEqual(dh1.pk);
        expect(tree.right.left.pk).toEqual(dh2.pk);
        expect(tree.right.right.pk).toEqual(dh3.pk);
        const n2 = (0, util_1.keypair)(src_1.crypt.hkdf(src_1.crypt.getSharedSecret(peer0.ephemeral.sk, dh1.pk), undefined, undefined, src_1.crypt.SK_LENGTH));
        const n3 = (0, util_1.keypair)(src_1.crypt.hkdf(src_1.crypt.getSharedSecret(dh2.sk, dh3.pk), undefined, undefined, src_1.crypt.SK_LENGTH));
        const n1 = (0, util_1.keypair)(src_1.crypt.hkdf(src_1.crypt.getSharedSecret(n2.sk, n3.pk), undefined, undefined, src_1.crypt.SK_LENGTH));
        expect(tree.pk).toEqual(n1.pk);
        const pks = util_1.Gen.collect(tree.members(), (n) => n.pk);
        expect(pks.length).toEqual(7);
        const reverse = serDeser(tree);
        expect(reverse.length).toEqual(7);
        expect(reverse).toBeInstanceOf(art_1.Node);
        expect(reverse.left).toBeInstanceOf(art_1.Node);
        expect(reverse.right).toBeInstanceOf(art_1.Node);
        expect(reverse.left.left.pk).toEqual(tree.left.left.pk);
        expect(reverse.left.right.pk).toEqual((0, util_1.keypair)((0, util_1.keyExchangeReceive)(peer1.identity, peer0.identity.pk, peer1.ephemeral, exchange.pk)).pk);
        expect(reverse.right.pk).toEqual(tree.right.pk);
        expect(reverse.pk).toEqual(tree.pk);
    });
    it('should split 3-tree at 3rd leaf', () => {
        const [peer0, peer1, peer2, peer3] = (0, test_utils_1.generate)(4);
        if (!peer0 || !peer1 || !peer2 || !peer3)
            throw new Error();
        const exchange = (0, util_1.keypair)();
        const peers = [
            new art_1.Me(peer0.identity, peer0.ephemeral),
            new art_1.Peer(peer1.identity.pk, peer1.ephemeral.pk, peer1.ephemeral_signature),
            new art_1.Peer(peer2.identity.pk, peer2.ephemeral.pk, peer2.ephemeral_signature),
        ];
        const tree = (0, art_1.createTree)(peers, {
            identity: peer0.identity,
            exchange,
        });
        expect(tree.length).toEqual(5);
        const pks = util_1.Gen.collect(tree.members(), (n) => n.pk);
        expect(pks.length).toEqual(5);
        const reverse = serDeser(tree);
        expect(reverse.length).toEqual(5);
        const exchange_new = (0, util_1.keypair)();
        expect(tree.split(tree.right.pk, new art_1.Peer(peer3.identity.pk, peer3.ephemeral.pk, peer3.ephemeral_signature), {
            identity: peer0.identity,
            exchange: exchange_new,
        })).toBe(true);
        expect(tree.length).toEqual(7);
        expect(tree.left).toBeInstanceOf(art_1.Node);
        expect(tree.right).toBeInstanceOf(art_1.Node);
        expect(tree.left.left).toBeInstanceOf(art_1.Leaf);
        expect(tree.left.right).toBeInstanceOf(art_1.Leaf);
        expect(tree.right.left).toBeInstanceOf(art_1.Leaf);
        expect(tree.right.right).toBeInstanceOf(art_1.Leaf);
        const dh1 = (0, util_1.keypair)((0, util_1.keyExchangeSend)(peer0.identity, peer1.identity.pk, exchange, peer1.ephemeral.pk));
        const dh2 = (0, util_1.keypair)((0, util_1.keyExchangeSend)(peer0.identity, peer2.identity.pk, exchange, peer2.ephemeral.pk));
        const dh3 = (0, util_1.keypair)((0, util_1.keyExchangeSend)(peer0.identity, peer3.identity.pk, exchange_new, peer3.ephemeral.pk));
        expect(tree.left.left.pk).toEqual(peer0.ephemeral.pk);
        expect(tree.left.right.pk).toEqual(dh1.pk);
        expect(tree.right.left.pk).toEqual(dh2.pk);
        expect(tree.right.right.pk).toEqual(dh3.pk);
        const n2 = (0, util_1.keypair)(src_1.crypt.hkdf(src_1.crypt.getSharedSecret(peer0.ephemeral.sk, dh1.pk), undefined, undefined, src_1.crypt.SK_LENGTH));
        const n3 = (0, util_1.keypair)(src_1.crypt.hkdf(src_1.crypt.getSharedSecret(dh2.sk, dh3.pk), undefined, undefined, src_1.crypt.SK_LENGTH));
        const n1 = (0, util_1.keypair)(src_1.crypt.hkdf(src_1.crypt.getSharedSecret(n2.sk, n3.pk), undefined, undefined, src_1.crypt.SK_LENGTH));
        expect(tree.pk).toEqual(n1.pk);
        expect(reverse.split(reverse.right.pk, new art_1.Peer(peer3.identity.pk, peer3.ephemeral.pk, peer3.ephemeral_signature), [dh3.pk, n3.pk, n1.pk])).toBe(true);
        expect(reverse.length).toEqual(7);
        expect(reverse.left).toBeInstanceOf(art_1.Node);
        expect(reverse.right).toBeInstanceOf(art_1.Node);
        expect(reverse.left.left).toBeInstanceOf(art_1.Leaf);
        expect(reverse.left.right).toBeInstanceOf(art_1.Leaf);
        expect(reverse.right.left).toBeInstanceOf(art_1.Leaf);
        expect(reverse.right.right).toBeInstanceOf(art_1.Leaf);
        expect(reverse.left.left.pk).toEqual(peer0.ephemeral.pk);
        expect(reverse.left.right.pk).toEqual(dh1.pk);
        expect(reverse.right.left.pk).toEqual(dh2.pk);
        expect(reverse.right.right.pk).toEqual(dh3.pk);
        expect(reverse.pk).toEqual(n1.pk);
        // expect((reverse.left as Node).left.pk).toEqual((tree.left as Node).left.pk);
        // expect((reverse.left as Node).right.pk).toEqual(
        //   keypair(keyExchangeReceive(peer1.identity, peer0.identity.pk, peer1.ephemeral, exchange.pk)).pk
        // );
        // expect(reverse.right.pk).toEqual(tree.right.pk);
        expect(reverse.pk).toEqual(tree.pk);
    });
});
